(* Struktur Env.sml implementiert Signatur Env.sig *)
(* Erklaerung siehe Env.sig                        *)

structure Env :> Env =
struct 

    open Data

    (* Datenstruktur env ist ein record bestehend aus 6 Komponenten *)
    type env = { tc : string -> ty ,                 (*  Const-Typumgebung      *)
                 tf : string -> ty ,                 (*  Var - Typumgebung      *)
		 clist : string list ,               (*  Const List             *)
		 flist : string list,                (*  Free Var  - Int List   *)
		 blist : string list,                (*  Bound Var - Int List   *) 
		 tlist : string list }               (*  Type List              *)
	    

(* --------------------------------------------------------------------*)
(* |                   BEGIN Hilfsprozeduren                          |*)
(* --------------------------------------------------------------------*)


    (* [debug_print]
     * DEBUGGING
     *)
    fun debug_print xs =  ("\n[ " ^ (foldl (fn (s,aus) => (aus ^ ", " ^ s) ) "" xs ) ^ " ]\n")

    
    (* [mapstring2bool]
     * generiert aus Typenliste eine Funktion
     * map : string -> bool mit
     * map[typ] = typ ist in typliste
     *)
    fun mapstring2bool (xs:string list) =
	let
	    val init = (fn s:string => false)
	    val map = foldl (fn (x,f) =>( fn s:string => if (s=x) then true else f s ) ) init xs
	in 
	    map
	end


    (* [mapstring2int]
     * generiert aus einer string list xs eine Funktion map
     * mit map[s] = position von s in liste xs
     *)
    fun mapstring2int (xs:string list) = 
	let
	    val (_,map) = 
		(foldl (fn (x,(n,f)) => 
			if (( f x ; false ) handle Error _ => true )
			    then (n+1, fn v => if v = x then n else f v) 
			else (n,f) ) (0,( fn v => raise Error ("Variable " ^ v 
							       ^ " muss zunaechst" 
							       ^ " deklariert werden!\n") 
					 )
				      ) xs 
		 )
	in
	    map
	end

    (* [mapint2string]
     * generiert aus einer string list xs eine Funktion map
     * mit map[i] = string an position i in liste xs
     *)
    fun mapint2string (xs:string list) = 
	(fn i => List.nth((xs,i)) handle Subscript => "" )


  
   (* [concat]
    * konkateniert zwei Stringlisten, allerdings duerfen positionen der alten liste
    * nicht veraendert werden und doppelauftreten nicht vorkommen
    *)
(*
    fun concat (xs:string list) nil = xs
      | concat nil ys = ys
      | concat xs (y::ys) = (concat xs ys)@( if (List.exists( fn s => s = y) xs) then nil else [y]  )
*)

    (* [next]
     * gibt zu einem String den numerisch folgenden Nachfolger
     * Bsp : x --> x0  oder  x2 --> x3 usw.
     *)
    fun digit_pos xs = foldl ( fn (c,n) => if Char.isDigit(c) then n else n+1 ) 0 xs 
    fun next s = 
	let
	    val ss = explode(s)
	    val pos = digit_pos ss
	    val prefix = String.substring(s, 0, pos)
	    val postfix = String.substring(s,pos, String.size(s) - pos)
	    val newpostfix = case Int.fromString(postfix) of
		               NONE   => "0"
			     | SOME i => Int.toString (i+1)
	in
	    prefix^newpostfix
	end


    (* [rename]
     * nimmt newxs neue string list
     * und (cs,ys,zs) alte string listen
     * prueft fuer jedes Element x aus newxs, ob
     * x in cs, ys oder zs vorkommt
     *    wenn ja   => wird x mihilfe von next umbenannt und erneut ueberprueft
     *    wenn nein => wird x uebernommen
     * Bei jedem Renaming von x muss ebenfalls ueberprueft werden, ob
     * next(x) nicht bereits Element der ueberarbeiteten Liste ist
     *)
(*  
  fun rename' nil    (cs, ys, zs)   resxs = resxs
      | rename'(x::xs) (cs, ys, zs)   resxs = if ( (List.exists( fn s => s = x ) ys) orelse 
						   (List.exists( fn s => s = x ) zs) orelse
						   (List.exists( fn s => s = x ) cs) orelse
						   (List.exists( fn s => s = x ) resxs )    ) 
					     then rename' (next(x)::xs) (cs, ys, zs) resxs
					     else rename' xs (cs, ys, zs)  (resxs @ [x])
    fun rename newxs (cs, ys, zs) = rename' newxs (cs, ys, zs) nil
*)

    fun extend bmap nil = bmap
      | extend bmap (x::xs) = if (List.exists ( fn s => s = x ) bmap) 
				  then extend bmap xs 
			      else extend (bmap @ [x]) xs 


(* --------------------------------------------------------------------*)
(* |                   ENDE Hilfsprozeduren                           |*)
(* --------------------------------------------------------------------*)


(* --------------------------------------------------------------------*)
(* |                   BEGIN Hauptmethoden                            |*)
(* --------------------------------------------------------------------*)


    (* [empty]
     * legt initiale Variablen/Konstanten Typumgebung,
     * leere Free/Bound Variablen Listen und 
     * initiale Typliste ["bool"] an 
     *)
    fun empty() = 
	let 
	    val init_typeenv_consts : string -> ty  = 
		fn x:string => if (x="TRUE") orelse (x="FALSE") 
				   then (TC "bool")
			       else raise Error ("Constante " ^ x 
						 ^ " muss zunaechst deklariert werden!\n")
	    val init_typeenv_freevars : string -> ty  = 
		fn x:string => raise Error ("Variable " ^ x 
					    ^ " muss zunaechst deklariert werden!\n")
	    val init_types   = ["bool"]        
	in 
	    { tc = init_typeenv_consts ,
	      tf = init_typeenv_freevars , 
	      clist = ["TRUE","FALSE"] ,                 
	      flist = nil ,
	      blist = nil ,
	      tlist = init_types }
	end
    
	      

    (* [new_type]
     * Erweitert die Typliste um den Typ <ntype> 
     *)
     fun new_type( oldenv:env, ntype:string) = 
	 let 
	     val { tc = ctypeenv, 
		   tf = ftypeenv,
		   clist = cmap,
		   flist = fmap,
		   blist = bmap,
		   tlist = typelist } = oldenv
	 in
	     (* Doppelauftreten werden vermieden *)
	     if (List.exists (fn s => s=ntype) typelist) 
		 then 
		     oldenv
	         else
		  { tc = ctypeenv, 
		    tf = ftypeenv,
		    clist = cmap,
		    flist = fmap,
		    blist = bmap,
		    tlist = (ntype::typelist) }
	 end
     
    
     (* [update_ctypeenv]
      * einer neuen Constanten <c> wird ein Typ t zugeordnet
      *)
     fun update_ctypeenv(oldenv:env,c:string,t:ty) = 
	 let
	     val { tc = ctypeenv, 
		   tf = ftypeenv,
		   clist = cmap,
		   flist = fmap,
		   blist = bmap,
		   tlist = typelist } = oldenv

	     (* prueft ob alle Basistypen in t ueberhaupt deklariert wurde *)
	     val typmap = mapstring2bool typelist
	     val _ = if (Typecheck.checktypetree t typmap) then () 
		     else raise Error ("Some Base Types in " ^ c ^ " are unknown \n")

	     val known =    (  (ctypeenv c; true ) handle Error _ => false   ) orelse
		            (  (ftypeenv c; true ) handle Error _ => false   )

	     val newctypeenv = (fn s => if s=c then t else ctypeenv s)
	 in
	     (* known=true  <=> Konstante c bereits deklariert wurde *)
	     if (known) then 
		 raise Error ("Konstante/Variable " ^c^ " wurde bereits deklariert\n")
	     else { tc = newctypeenv, 
		    tf = ftypeenv,
		    clist = cmap,
		    flist = fmap,
		    blist = bmap,
		    tlist = typelist }
	 end

      (* [update_ftypeenv]
      * einer neuen Variablen <v> wird ein Typ t zugeordnet
      *)
     fun update_ftypeenv(oldenv:env,v:string,t:ty) = 
	 let
	     val { tc = ctypeenv, 
		   tf = ftypeenv,
		   clist = cmap,
		   flist = fmap,
		   blist = bmap,
		   tlist = typelist } = oldenv

	     (* prueft ob alle Basistypen in t ueberhaupt deklariert wurde *)
	     val typmap = mapstring2bool typelist
	     val _ = if (Typecheck.checktypetree t typmap) then () 
		     else raise Error ("Some Base Types in " ^ v ^ " are unknown \n")

	     val known =    (  (ftypeenv v; true ) handle Error _ => false   ) orelse
		            (  (ctypeenv v; true ) handle Error _ => false   )
	     val newftypeenv = (fn s => if s=v then t else ftypeenv s)
	 in
	     (* known=true  <=> Konstante v bzw. Freie Variable v bereits deklariert wurde *)
	     if (known) then 
		 raise Error ("Variable/Konstante " ^v^ " wurde bereits deklariert\n")
	     else { tc = ctypeenv, 
		    tf = newftypeenv,
		    clist = cmap,
		    flist = fmap,
		    blist = bmap,
		    tlist = typelist }
	 end
 


     (* [update_constmaplist]
      * der Constmapliste werden die Konstanten der Liste ms hinzugefuegt 
      *)
     fun update_constmaplist(oldenv:env, ms:string list) = 
	 let 
	     val { tc = ctypeenv, 
		   tf = ftypeenv,
		   clist = cmap,
		   flist = fmap,
		   blist = bmap,
		   tlist = typelist } = oldenv
	 in
	     { tc = ctypeenv, 
	       tf = ftypeenv,
	       clist = extend cmap ms,
	       flist = fmap,
	       blist = bmap,
	       tlist = typelist }
	 end


     (* [update_freemaplist]
      * der Freemapliste werden die Variablen der Liste ms hinzugefuegt 
      *)
     fun update_freemaplist(oldenv:env, ms:string list) = 
	 let 
	     val { tc = ctypeenv, 
		   tf = ftypeenv,
		   clist = cmap,
		   flist = fmap,
		   blist = bmap,
		   tlist = typelist } = oldenv
	 in
	     { tc = ctypeenv, 
	       tf = ftypeenv,
	       clist = cmap,
	       flist = extend fmap ms,
	       blist = bmap,
	       tlist = typelist }
	 end
		
  
     (* [update_bndmaplist]
      * da neue gebundene Variablen hinzugefuegt werden,
      * kann es zu Konflikten mit bereits vorhandenen
      * Konstanten, freien Variablen oder gebundenen Variablen kommen
      * => Umbenennung der alten gebundenen Variablen 
      *)
     fun update_bndmaplist(oldenv:env, ms:string list) =
	 let 
	    (* DEBUG val _ = foldl (fn (s,_) => print (s ^ "\n")) () ms *) 
	     val { tc = ctypeenv, 
		   tf = ftypeenv,
		   clist = cmap,
		   flist = fmap,
		   blist = bmap,
		   tlist = typelist } = oldenv

	 (*    val newbmap = rename bmap ( cmap, fmap, ms ) *)
 	 (* !!! Die alten Geb. Var. werden veraendert, nicht die neuen !!! *)
	     val newbmap = extend bmap ms
	 in
	     { tc = ctypeenv, 
	       tf = ftypeenv,
	       clist = cmap,
	       flist = fmap,
	     (*  blist = newbmap @ ms , *)
	       blist = newbmap,
	       tlist = typelist }
	 end

(* [renameBndVars] 
 * nimmt Liste von Gebundenen Variablen ( Integers ! )
 * transformiert diese in eine Stringliste und 
 * nennt die Schnittmenge dieser Liste mit der bmap liste
 * um !
 *)  

     fun renameBndVars' nil f = f
       | renameBndVars' (b::bs) f = renameBndVars' bs ( fn s=> if s=b then next b else f s )
     fun renameBndVars(oldenv:env, is: int list) =
	 let
	      val { tc = ctypeenv, 
		   tf = ftypeenv,
		   clist = cmap,
		   flist = fmap,
		   blist = bmap,
		   tlist = typelist } = oldenv

	      val bndmapi2s = mapint2string(bmap)
	      val slist = List.map bndmapi2s is
	      val f = renameBndVars' slist (fn s:string => s)
	      val newbmap = List.map f bmap
	 in
	      { tc = ctypeenv, 
	       tf = ftypeenv,
	       clist = cmap,
	       flist = fmap,
	       blist = newbmap,
	       tlist = typelist }
	 end

(* [freeVarInt]
 * searches the next free name, that isn't used yet in freemap or constmap
 * is called by Intro All /Elim Exists
 *) 


     fun freeVarIntro (oldenv:env, i:int) = 
    	 let
	      val { tc = ctypeenv, 
		   tf = ftypeenv,
		   clist = cmap,
		   flist = fmap,
		   blist = bmap,
		   tlist = typelist } = oldenv

	      val bndmapi2s = mapint2string(bmap)
	      val oldname = bndmapi2s i 
		  
	      fun search_newname s = 
		  if (List.exists (fn z=>z=s) cmap) orelse
		     (List.exists (fn z=>z=s) fmap) 
 		      then search_newname (next s) else s
	 in
	     search_newname oldname
	 end

(* --------------------------------------------------------------------*)
(* |                   ENDE Hauptmethoden                             |*)
(* --------------------------------------------------------------------*)


(* --------------------------------------------------------------------*)
(* |                   BEGIN Instanziierungsmethoden                  |*)
(* --------------------------------------------------------------------*)
  (* [assert_free_bnd]
      * Stellt sicher, dass bndmap und (freemap, constmap) invariant disjunkt sind 
      *)
 (*    fun assert_free_bnd(oldenv:env) = 
	 let
	     val { tc = ctypeenv, 
		   tf = ftypeenv,
		   clist = cmap,
		   flist = fmap,
		   blist = bmap,
		   tlist = typelist } = oldenv

	     val newbmap = rename bmap (cmap, fmap, nil)
	 in 
	     { tc = ctypeenv, 
	       tf = ftypeenv,
	       clist = cmap,
	       flist = fmap,
	       blist = newbmap,
	       tlist = typelist }
	 end
*)
 

(*
    
     (* [gen_instance_env] 
      * gebe neue Env zurueck  ( Im Falle einer Instanziierung )
      * mit:
      * - BndVarMap der aktuellen/oldenv Umgebung
      * - ConstMap, FreeVarMap, Typelist, Typmap der gammaenv
      *)
      fun gen_instance_env( oldenv, gammaenv) = 
	 let
	     val oldbmap = #blist oldenv

	     val { tc = gammactypeenv, 
		   tf = gammaftypeenv,
		   clist = gammacmap,
		   flist = gammafmap,
		   blist = _ ,
		   tlist = gammatypelist } = gammaenv
	 in
	     { tc = gammactypeenv, 
	       tf = gammaftypeenv,
	       clist = gammacmap,
	       flist = gammafmap,
	       blist = oldbmap,
	       tlist = gammatypelist }
	 end
 *)    

     (* [back_instance_env]
      * Die aktuelle Environment wird zurueckkopiert und dabei nur die neue Bnd Map Liste
      * uebernommen, allerdings muss nun sichergestellt werden, dass in der Instanziierung
      * keine gebundenen Variablen verwendet wurden, die in der aktuellen Umgebung erst 
      * frei geworden sind. Deshalb der assert_free check !
      *)
      fun back_instance_env( oldenv:env, instenv:env ) = 
	  let
	      val { tc = ctypeenv, 
 		    tf = ftypeenv,
		    clist = cmap,
		    flist = fmap,
		    blist = bmap,
		    tlist = typelist } = oldenv

	      val instbmap = #blist instenv
	  
	      val newenv = { tc = ctypeenv, 
			     tf = ftypeenv,
			     clist = cmap,
			     flist = fmap,
			     blist = instbmap,
			     tlist = typelist }
	  in
	    (*  assert_free_bnd(newenv) *)
	      newenv
	  end
      
      
(* --------------------------------------------------------------------*)
(* |                   ENDE Instanziierungsmethoden                   |*)
(* --------------------------------------------------------------------*)



(* --------------------------------------------------------------------*)
(* |             BEGIN MAPPING Generatoren Methoden                   |*)
(* --------------------------------------------------------------------*)
  
  
  (* Alle folgenden Methoden generieren aus den entsprechenden
   * Listen Mapping Funktionen
   *)
      fun freemaps2i ( oldenv:env ) = let
				      val fmap = #flist oldenv
				  in mapstring2int fmap
				  end
			      
      fun freemapi2s ( oldenv:env ) = let
				      val fmap  = #flist oldenv
				  in mapint2string fmap
				  end
	      
      fun bndmaps2i ( oldenv:env ) = let
				     val bmap = #blist oldenv
				 in mapstring2int bmap
				 end
      fun bndmapi2s ( oldenv:env ) = let
				     val bmap = #blist oldenv
				 in mapint2string bmap
				 end
			     

      fun constmapi2s (oldenv:env) = let
					 val cmap = #clist oldenv
				     in mapint2string cmap
				     end

      fun constmaps2i (oldenv : env) = let
					   val cmap = #clist oldenv
				       in mapstring2int cmap
				       end

      fun typmaps2b (oldenv:env) = let
				   val typelist = #tlist oldenv
			       in mapstring2bool typelist
			       end
			
      (* [typi]
       * gibt den Typ der Variable mit der Nummer i zurueck
       *)
      fun typi( oldenv:env, x:int) = let 
				     val ftypeenv = #tf oldenv
				     val fmap = #flist oldenv
				     val variable = (mapint2string fmap) x 
				 in 
				     ftypeenv variable
				 end

      (* [typc]
       * gibt den Typ der Konstante c zurueck
       *)		     
      fun typc( oldenv:env, c:string) = let 
					val ctypeenv = #tc oldenv
				    in 
					ctypeenv c
				    end
	
      (* [typf]
       * gibt den Typ der Variable v zurueck
       *)		     
      fun typf( oldenv:env, v:string) = let 
					val ftypeenv = #tf oldenv
				    in 
					ftypeenv v 
				    end
				
(* --------------------------------------------------------------------*)
(* |                   ENDE MAPPING Generatoren Methoden              |*)
(* --------------------------------------------------------------------*)

(* --------------------------------------------------------------------*)
(* |                   Verschiedene Methoden                          |*)
(* --------------------------------------------------------------------*)

      (* [nr_consts]
       * gibt die anzahl der deklarierten Konstanten zurueck 
       *)
      fun nr_consts (oldenv:env) = let
				   val cmap = #clist oldenv
			       in List.length(cmap)
			       end
			   
      (* [nr_freevars]
       * gibt die anzahl der deklarierten freien Variablen zurueck 
       *)
      fun nr_freevars (oldenv:env) = let
				     val fmap = #flist oldenv
				 in List.length(fmap)
				 end

      (* [nr_bndvars]
       * gibt die Anzahl der bekannten gebundenen Variablen zurueck
       *)
      fun nr_bndvars (oldenv:env) = let
				    val bmap = #blist oldenv
				in List.length(bmap)
				end
			  
			  
end (* of struct *)



