(* Equality.sml
 * implements the Equality Extension Rules :
 * @ Leibniz Extension :        A(:i) = B(:i)  --->  \/p:i->bool. p A => p B
 * @ Boolean Extensionality :   A(:bool) = B(:bool) ---> A => B (1/2)
 *                                                       B => A (2/2)
 * @ Functional Extensionality: 
 *                    (1)        A(:a->b) = B(:a->b)  ---> \/x:a. A x = B x 
 *		      (2)   (\x:a. A ) = (\x:a. B )   ---> \/x:a. A = B
 *
 * ----------------------------------------------------
 * | !! Equality.sml returns ( newterm , newgoal ) !! |
 * ----------------------------------------------------
 * 
*)                                 


local open Data Goal in

fun main (term, goal) = process term goal

and process (EQ(eqleft,eqright)) goal = 
    let

	 val cmap = Goal.typc(goal)
	 val fmap = Goal.typi(goal)    (* Free Var Map *)
	 val tmap = Goal.typemap(goal) (* Typlisten Map *)

	 (* Hole Typ von subterm a *)
	 val typ_a = Typecheck.get_type eqleft (cmap,fmap,tmap)
	 (* Hole Typ von subterm b *)
	 val typ_b = Typecheck.get_type eqright (cmap,fmap,tmap)

	 val _ = ( if (typ_a <> typ_b) 
		     then raise Error("Different Types in Equality !\n")
		   else () )


	(* [leibniz]
	 * takes basetype i and returns the expanded LeibnizEquality 
	 *        ... => A:i = B:i
	 * ---->  ... => \/p:i->bool. p A => p B
	 *)	     
	 fun leibniz(i) = 
	     let 
		 (* Fuege Geb.Variable <p> der Umgebung hinzu *)
		 val goal' = Goal.update_bndvar(goal, "p") 
		 val nextbndnr = Goal.nr_bndvars(goal') - 1
		 val newterm = ( ALL((TA(i, TC "bool"), nextbndnr), 
				      IMP(A(V 0 , (ETT.up 0 eqleft 1)),
					  A(V 0 , (ETT.up 0 eqright 1) ))) 
				 )

	     in
		  (newterm, goal')
	     end

	 (* [funext]
	 * takes type alpha and returns the expanded Functional
	 * extended Term of "right"
	 *          ... => A(:alpha->beta) = B(:alpha->beta)
	 * ---->    ... => \/z:alpha. A z = B z
	 *)	     
	 fun funext(alpha) = 
	     let 
	 	 (* Fuege Geb.Variable <z> der Umgebung hinzu *)
		 val goal' = Goal.update_bndvar(goal, "z") 
		 val nextbndnr = Goal.nr_bndvars(goal') - 1
		 val newterm = ( ALL((alpha, nextbndnr), 
				      EQ(A( (ETT.up 0 eqleft 1), V 0),
					 A( (ETT.up 0 eqright 1), V 0)))
				 )
     
		
	     in
		 (newterm, goal')
	     end

	 (* [lamext]
	 * substitutes two lambda terms in equality by an ALL expression
	 *          ... => (\x:alpha. A) = (\x:alpha. B)
	 * ---->    ... => \/x:alpha. A  =  B 
	 *)	     
	 fun lamext() = 
	     let 
		 val (alpha, nr) = ( case (eqleft) of
				    (L((t,i),_)) => (t,i)
				  | _ => raise Error("Equality.sml Internal Nr. 4 \n") )
		 val (l, r) = ( case (eqleft,eqright) of
			       ( L(_,l) , L(_,r)) => (l,r)
			     | _ => raise Error("Equality.sml Internal Nr. 5 \n") )

		 val newterm = ( ALL((alpha, nr), EQ(l,r) ))
     
	     in
		 (newterm, goal)
	     end


	 fun funextORlam(alpha) =
	     case (eqleft,eqright) of
		 (L(_) , L(_)) => lamext()
	       | _             => funext(alpha) 
		      

	 (* [boolext]
	 * performs the Boolean Extensionality Rule
	 * Bsp.:
	 *          C => A:bool = B:bool
	 * ---->    C => (A => B) & (B => A )
	 *)	
	 fun boolext() = 
	     let
		val term1  = IMP(eqleft, eqright)
		val term2  = IMP(eqright, eqleft)
		val newterm = AND(term1, term2)
	     in
		 (newterm, goal)
	     end



	 (* Waehle wie die Gleichheit ersetzt werden soll *)
	 val (newterm, newgoal) = 
	     (  case typ_a of 
		    (TC "bool")      => boolext()
		  | (i as (TC _ ))   => leibniz(i)
		  | (TA(alpha,beta)) => funextORlam(alpha)
	     )

     in
	 (newterm, newgoal)
     end

| process t goal = (t,goal)

end (* of local open *)