structure Gamma :> Gamma =
struct 

    open Data
    (* Ein Knoten kann (nicht markiert), entdeckt oder geschl. sein *)
    datatype modus = nonmarked | discovered | closed   

    (* Ein Knoten im Graphen ist entweder Gamma oder Delta Variable *)
    datatype nodevalue = GAMMA_VAR of int | DELTA_VAR of int
    (* State Merkfeld, Information Gamma/Delta int, Nachfolgeknoten *)

    (* Graph Datenstruktur *)
    datatype state = G of modus * nodevalue * node list       
    withtype node = state ref                           (* Knoten ist eine Zeiger auf state *)
    type graph = node list                              (* Graph ist eine Liste von Knoten *)

    (* Interne exceptions *)
    exception cycle
    exception notfound
    exception neverhappen

    type gamma = graph  (* Die Gamma/Delta Variablen werden in einem Graph
			 * verwaltet 
			 * Der Abhaengigkeitstest erfolgt durch Cyclustest!
			 *)


    (* ----------------------------------------------------------------- *)
    (* |     BEGIN        Hilfsprozeduren Graphbearbeitung             | *)
    (* ----------------------------------------------------------------- *)
	
    fun cons x gs  = ref (G(nonmarked, x ,gs))
    fun assign (g as ref(G(m,x,_))) gs = g := G(m,x,gs)
	
    fun create_edge (k1 as ref(G(m1,x1,g1))) k2 = (k1 := G(m1,x1, (k2::g1)))
    
    fun remove_edge (k1 as ref(G(m1,x1,g1))) k2 = 
	let 
	    fun filter (nil, element) result     = result
	      | filter ((x::xs), element) result = 
		if (x = element) then filter (xs, element) result
		else filter (xs, element) (x::result)
	in
	    k1 := G(m1, x1, filter (g1, k2) nil)
	end


    (* sucht im Graph den Knoten mit dem Wert x *)
    fun find_node nil x = raise notfound
      | find_node ( (g as (ref(G(_, v, _))))::gs ) x = 
	if (x=v) then g else find_node gs x
    


    fun discover (g as ref(G(_,x,cs))) = (g:=G(discovered,x,cs))
    fun close (g as ref(G(_,x,cs))) = (g:=G(closed,x,cs))
	
    fun  reset nil = ()
      |  reset ((g as (ref (G(nonmarked, x,gs))))::xr) = reset xr 
      |  reset ((g as (ref (G(_, x, gs))))::xr) = (g:=G(nonmarked, x, gs); reset (gs@xr) )
	
	
    (* Zyklustest *)
    (* DFS Algorithmus *)
	
    fun cyclic g = ( reset g ; cyclic' g )
    and cyclic' nil = false
      | cyclic' ((g as ref(G(closed,_,_)))::gs) = cyclic gs
      | cyclic' ((g as ref(G(nonmarked,x,cs)))::gs) =( ( discover g ; 
							dfs [g] ; 
							close g ;
							cyclic' gs ) handle cycle => true )
      | cyclic' _ = raise neverhappen
	
    and dfs nil = ()
      | dfs ((s as ref(G(discovered, x, nil)))::stack ) = ( close s ; dfs stack ) (* pop s and backtrack *)
      | dfs ((s as ref(G(discovered, x, childs )))::stack) = 
	( case (nextnonmarked (childs)) of
	    nil => ( close s; dfs stack )
	  | (c::nil) =>( discover c; dfs (c::s::stack) )
	  | _ => raise neverhappen
		)
      | dfs _ = raise neverhappen

    and nextnonmarked (nil) = nil 
      | nextnonmarked ( (c as ref(G(nonmarked,_,_)))::cs) = [c]
      | nextnonmarked ( (c as ref(G(discovered,_,_)))::cs) = raise cycle
      | nextnonmarked ( (c as ref(G(closed,_,_)))::cs) = nextnonmarked cs
	
	
    (* ----------------------------------------------------------------- *)
    (* |     ENDE         Hilfsprozeduren Graphbearbeitung             | *)
    (* ----------------------------------------------------------------- *)


    (* ----------------------------------------------------------------- *)
    (* |     BEGIN        Hilfsprozeduren Gamma Routinen               | *)
    (* ----------------------------------------------------------------- *)

		
	(* fuegt allen Gamma Knoten innerhalb des Graphens eine 
	 * Kante zum Delta Knoten hinzu 
	 *)

	fun delta_update graph =
	    let 
		val delta = hd(graph) (* da gerade hinzugefuegt *)
	    in
		foldl ( fn (node, _ ) => 
		       case (node) of
			   (ref (G(_, GAMMA_VAR(_),_))) => create_edge node delta
			 | _ => ()
		       )      
		        () (tl(graph))
	    end




    (* ----------------------------------------------------------------- *)
    (* |     BEGIN        Gamma Haupt Routinen                         | *)
    (* ----------------------------------------------------------------- *)

	

    (* Legt initialen Graph an : leer *)
    fun empty () = nil



    (* Fuegt Eine Gamma Varible in den Graph ein *)
    fun add_gamma (gs, i) = 
	let 
	    val new_node = cons (GAMMA_VAR i) nil
	    val new_graph = (new_node::gs)
	in
	    new_graph
	end
    
    
    (* Fuegt Eine Delta Varible in den Graph ein *)
    fun add_delta (gs, i) =
	let
	    val new_node = cons (DELTA_VAR i) nil
	    val new_graph = (new_node::gs)

	    val _ = delta_update(new_graph)
	in 
	    new_graph 
	end

    
    
    (* Prueft, ob eine Substitution zulaessig ist, sprich kein
     * Cyclus im Graph ist, nachdem alle Abhaengigkeitsedges hinzugefuegt
     * wurden
     *)
    
    fun substitute gs (i, terms ) =
	let
	    (* gamma_node ist der Knoten der Gammavariable i im Graph *)
	    val gamma_node = ( find_node gs (GAMMA_VAR i) ) 
	                      handle notfound => 
				 raise Error ("There is no Gamma Variable ?" ^ Int.toString(i) ^ "\n")

	    (* Liste mit allen Edges, die durch die Substitution bedingt
	     * dem Graphen hinzugefuegt werden
	     * Im Falle einer nicht erlaubten Substitution ( <=> Cyclus im Graph )
	     * werden all_new_edges wieder geloescht 
	     *)
	    val all_new_edges =
		foldl ( fn (t, edges) => 
		       ( case t of
			   (V i)    => ( (find_node gs (DELTA_VAR i) , gamma_node) :: edges )
			 |(GAMMA (_,i)) => ( (find_node gs (GAMMA_VAR i) , gamma_node) :: edges )
			 | t => raise Error (" Internal Error Gamma.sml; not only variables in list \n")
			       ) handle notfound => edges 
			     ) nil terms
	       

	    (* fuege nun diese edges dem Graph hinzu *)
	    val _ = List.app (fn (n1,n2) => create_edge n1 n2) all_new_edges

	    val zyklisch = cyclic gs
	in
	    (* Ist Graph nun zyklisch, loesche wieder alle Edges und gib false zurueck *)
	    if (zyklisch) 
		then ( List.app (fn (n1,n2) => remove_edge n1 n2) all_new_edges; false )
	        else true
	end
		


    (* counts the number of Gamma nodes in the graph *)
    fun nr_gammavars gs = 
	foldl ( fn (node,n) => 
	       case node of 
		   ( ref (G(_,GAMMA_VAR(_),_))) => n+1
		 | _ => n
		       ) 0 gs



		  
end (* of struct *)



