{
 open Lexing Parser;

 exception LexicalError of string * int * int (* (message, loc1, loc2) *)

 fun lexerError lexbuf s = 
     raise LexicalError (s, getLexemeStart lexbuf, getLexemeEnd lexbuf);

 val commentStart = ref 0;  (* Start of outermost comment being scanned *)
 
 fun commentNotClosed lexbuf =
     raise LexicalError ("Comment not terminated", 
                         !commentStart, getLexemeEnd lexbuf);
     
 val commentDepth = ref 0;  (* Current comment nesting *)

 (* Scan keywords as identifiers and use this function to distinguish them. *)
 (* If the set of keywords is large, use an auxiliary hashtable.            *)

fun keyword s = (* Keywords *)
     case s of
         "with"             => TWITH
 (* Hilfstoken *)
       | "type"             => TTYPE
       | "const"            => TCON 
       | "var"              => TVAR
       | "def"              => TDEF
       | "lemma"            => TLEMMA
       | "axiom"            => TAXIOM
       | "apply"            => TAPPLY
 (* Gleichheitsregeln *)
       | "rew"              => TREW
       | "pos"              => TPOS
       | "subst"            => TINSERT
       | "ref"              => TREF
       | "sym"              => TSYM
       | "trans"            => TTRANS
       | "beta"             => TBETA
       | "eta"              => TETA
       | "lam"              => TLAM
 (* ** Kommandos ** *)
       | "goal"             => TGOAL
       | "elim"             => TELIM
       | "remove"           => TREMOVE
       | "instance"         => TINSTANCE
       | "move"             => TMOVE
       | "shift"            => TSHIFT
       | "intro"            => TINTRO
       | "cut"              => TCUT
       | "simplify"         => TSIMPLIFY
       | "save"             => TSAVE
       | "help"             => THELP
       | "print"            => TPRINT
       | "labels"       => TSHOWLABELS
       | "latex"            => TLATEX
       | "undo"             => TUNDO
       | "reset"            => TRESET
       | "exit"             => TEXIT
       | "FALSE"            => TCONST "FALSE"
       | "TRUE"             => TCONST "TRUE"
       | "false"            => TCONST "FALSE"
       | "true"             => TCONST "TRUE"
       | _                  => TNAME s;
 
 }

rule Token = parse
    [` ` `\t` `\r`]     { Token lexbuf }
  | `~`?[`0`-`9`]+      { case Int.fromString (getLexeme lexbuf) of
                               NONE   => lexerError lexbuf "internal error"
               		     | SOME i => TNR i
                        }  
  | [`a`-`z``A`-`Z`][`a`-`z``A`-`Z``0`-`9` `_` `'`]*
                        { keyword (getLexeme lexbuf) }
  | "(*"                { commentStart := getLexemeStart lexbuf;
                          commentDepth := 1; 
                          SkipComment lexbuf; Token lexbuf }
  | "->"                { TDASHARROW }
  | `+`                 { TOR }                     
  | `&`                 { TAND }  
  | `~`                 { TNOT } 
  | "=>"                { TIMP }
  | "<=>"               { TAEQUI }
  | "="                 { TEQ }
  | `\\`                { TLAMBDA }                     
  | "\\/"               { TALL }
  | "/\\"               { TEXISTS }
  | `(`                 { TLPAR }
  | `)`                 { TRPAR }
  | `;`                 { TSEMI }
  | `:`                 { TDOUBLEPT }
  | `,`                 { TCOMMA }
  | `.`                 { TDOT }
  | `\n`                { TEOL }
  | eof                 { TEOF }
  | _                   { lexerError lexbuf "Illegal symbol in input" }

and SkipComment = parse
    "*)"                { commentDepth := !commentDepth - 1;  
                          if !commentDepth = 0 then ()
                          else SkipComment lexbuf 
                        } 
   | "(*"               { commentDepth := !commentDepth + 1; 
                          SkipComment lexbuf }
   | (eof | `\^Z`)      { commentNotClosed lexbuf }
   | _                  { SkipComment lexbuf }
;
